/*
 * Copyright (c) 1994 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Dlg/dlgfeature.h>
#include <Dlg/dlgfclass.h>
#include <Dlg/dlgcov.h>
#include <Dlg/dlglib.h>
#include <Dlg/dlgdb.h>
#include <Map/mapprim.h>
#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>
#include <stdio.h>
#include <stream.h>
#include <string.h>

/*****************************************************************************/

DlgFeature::DlgFeature (DlgFeatureClass* fclass, int id, MapFeatureType type) 
: MapFeature(fclass, id, type) {
  if (!attrlist())
    attrlist(new AttributeList());
  attrlist()->Append(new Attribute("id", new AttributeValue(id, AttributeValue::IntType)));
}

void DlgFeature::set_attributes(int major, int minor) {
    _major_code = major;
    _minor_code = minor;
  attrlist()->Append(new Attribute("major_code", new AttributeValue(major, AttributeValue::IntType)));
  attrlist()->Append(new Attribute("minor_code", new AttributeValue(minor, AttributeValue::IntType)));
}

short int DlgFeature::short_attr(char* attrname) {
    if (strcmp(attrname,"MAJOR_CODE")==0)
	return _major_code;
    else if (strcmp(attrname,"MINOR_CODE")==0)
	return _minor_code;
    else
	return MapFeature::short_attr(attrname);
}

long int DlgFeature::long_attr(char* attrname) {
    if (strcmp(attrname,"MAJOR_CODE")==0)
	return _major_code;
    else if (strcmp(attrname,"MINOR_CODE")==0)
	return _minor_code;
    else
	return MapFeature::long_attr(attrname);
}

boolean DlgFeature::definition(ostream& out) {
    out << " :id " << id();
    out << " :major_code " << _major_code << " :minor_code " << _minor_code;
    out << " :feature_desc " << "\"" << describe() << "\"";
    DlgFeatureClass* fc = (DlgFeatureClass*)feature_class();
    out << " :mapfc " << "\"" << fc->name() << "\"";
    DlgCoverage* cov = fc->coverage();
    out << " :mapcov " << "\"" << cov->name() << "\"";
    out << " :mapcov_desc " << "\"" << cov->description() << "\"";
    DlgLibrary* lib = cov->library();
    out << " :maplib " << "\"" << lib->name() << "\"";
    DlgDatabase* db = lib->database();
    out << " :mapdb " << "\"" << db->name() << "\"";
    out << " :mapdb_path " << "\"" << db->path() << "\"";
    return out.good();
}

const char* DlgFeature::describe() const {
    static char buffer[BUFSIZ];
    static char buf[255];
    char* typestring;
    switch (major_attribute()) {
    case 40: // water bodies
	switch (minor_attribute()) {
	case 100: typestring = "Perennial lake or pond"; break;
	case 102: typestring = "Intermittent lake or pond"; break;
	case 104: typestring = "Dry lake or pond"; break;
	case 105: typestring = "Alkali flat"; break;
	case 106: typestring = "Reservoir"; break;
	case 107: typestring = "Intermittent reservoir"; break;
	case 110: typestring = "Glacier or snowfield"; break;
	case 150: typestring = "Island"; break;
	case 199: typestring = "Area not in water body (remainder of map)"; break;
	default: 
	    sprintf(buf, "Unknown water bodies feature (%d)", minor_attribute());
	    typestring = buf;
	    break;
	}
	break;
    case 90: // political boundaries
	switch (minor_attribute()) {
	case 100: typestring = "Civil township, district, precint, or barrio"; break;
	case 101: typestring = "Incorporated city, village, town, borough, or hamlet"; break;
	case 197: typestring = "Canada"; break;
	case 198: typestring = "Mexico"; break;
	case 199: typestring = "Area outside a national boundary"; break;
	case 103: typestring = "National park, monument, lakeshore, parkway, battlefield, or recreation area"; break;
	case 104: typestring = "National forest or grassland"; break;
	case 105: typestring = "National wildlife refuge, game preserve, or fish hatchery"; break;
	case 106: typestring = "National scenic waterways or wilderness area"; break;
	case 107: typestring = "Indian reservation"; break;
	case 108: typestring = "Military reservation"; break;
	default: typestring = "Unknown political/administrative boundaries feature"; break;
	};
	break;
    case 91:
	sprintf(buf, "State FIPS code %d", minor_attribute());
	typestring = buf;
	break;
    case 92:
	sprintf(buf, "County or county equivalent FIPS code %d", minor_attribute());
	typestring = buf;
	break;
    case 102:
	sprintf(buf, "Interstate route %d", minor_attribute());
	typestring = buf;
	break;
    case 103:
	sprintf(buf, "U.S. route %d", minor_attribute());
	typestring = buf;
	break;
    case 104:
	sprintf(buf, "State route %d", minor_attribute());
	typestring = buf;
	break;
    case 290:
	switch (minor_attribute()) {
	case 2017: typestring = "Continental Divide"; break;
	case 3001: typestring = "River/stream (double line, shoreline)"; break;
	case 3002: typestring = "River/stream (double line, centerline)"; break;
	case 3003: typestring = "River/stream (single line), perennial, length < 20 km, or < l2 mi"; break;
	case 3004: typestring = "River/stream (single line), perennial, length 20-< 30 km, or l2-< l9 mi"; break;
	case 3005: typestring = "River/stream (single line), perennial, length 30-< 40 km, or l9-< 25 mi"; break;
	case 3006: typestring = "River/stream (single line), perennial, length 40-< 50 km, or 25-< 3l mi"; break;
	case 3007: typestring = "River/stream (single line), perennial, length 50-< 60 km, or 3l-< 37 mi"; break;
	case 3008: typestring = "River/stream (single line), perennial, length 60-< 80 km, or 37-< 50 mi"; break;
	case 3009: typestring = "River/stream (single line), perennial, length 80-< l00 km, or 50-< 62 mi"; break;
	case 3010: typestring = "River/stream (single line), perennial, length l00-< l25 km, or 62-< 78 mi"; break;
	case 3011: typestring = "River/stream (single line), perennial, length l25-< l50 km, or 78-< 93 mi"; break;
	case 3012: typestring = "River/stream (single line), perennial, length l50-< 200 km, or 93-< l24 mi"; break;
	case 3013: typestring = "River/stream (single line), perennial, length 200-< 250 km, or l24-< l55 mi"; break;
	case 3014: typestring = "River/stream (single line), perennial, length 250-< 300 km, or l55-< l86 mi"; break;
	case 3015: typestring = "River/stream (single line), perennial, length 300-< 350 km, or l86-< 2l7 mi"; break;
	case 3016: typestring = "River/stream (single line), perennial, length 350+ km, or 2l7+ mi"; break;
	case 3017: typestring = "River/stream (single line), intermittent, length < 20 km, or < l2 mi"; break;
	case 3018: typestring = "River/stream (single line), intermittent, length 20-< 30 km, or l2-< l9 mi"; break;
	case 3019: typestring = "River/stream (single line), intermittent, length 30-< 40 km, or l9-< 25 mi"; break;
	case 3020: typestring = "River/stream (single line), intermittent, length 40-< 50 km, or 25-< 3l mi"; break;
	case 3021: typestring = "River/stream (single line), intermittent, length 50-< 60 km, or 3l-< 37 mi"; break;
	case 3022: typestring = "River/stream (single line), intermittent, length 60-< 80 km, or 37-< 50 mi"; break;
	case 3023: typestring = "River/stream (single line), intermittent, length 80-< l00 km, or 50-< 62 mi"; break;
	case 3024: typestring = "River/stream (single line), intermittent, length l00-< l25 km, or 62-< 78 mi"; break;
	case 3025: typestring = "River/stream (single line), intermittent, length l25-< l50 km, or 78-< 93 mi"; break;
	case 3026: typestring = "River/stream (single line), intermittent, length l50-< 200 km, or 93-< l24 mi"; break;
	case 3027: typestring = "River/stream (single line), intermittent, length 200-< 250 km, or l24-< l55 mi"; break;
	case 3028: typestring = "River/stream (single line), intermittent, length 250-< 300 km, or l55-< l86 mi"; break;
	case 3029: typestring = "River/stream (single line), intermittent, length 300-< 350 km, or l86-< 2l7 mi"; break;
	case 3030: typestring = "River/stream (single line), intermittent, length 350+ km, or 2l7+ mi"; break;
	case 3035: typestring = "River/stream, centerline in water body, perennial, length < 2 km, or < l mi"; break;
	case 3036: typestring = "River/stream, centerline in water body, perennial, length 2-< 4 km, or l-< 2 mi"; break;
	case 3037: typestring = "River/stream, centerline in water body, perennial, length 4-< 6 km, or 2-< 4 mi"; break;
	case 3038: typestring = "River/stream, centerline in water body, perennial, length 6-< 8 km, or 4-< 5 mi"; break;
	case 3039: typestring = "River/stream, centerline in water body, perennial, length 8-< l0 km, or 5-< 6 mi"; break;
	case 3040: typestring = "River/stream, centerline in water body, perennial, length l0-< l5 km, or 6-< 9 mi"; break;
	case 3041: typestring = "River/stream, centerline in water body, perennial, length l5-< 20 km, or 90-< l2 mi"; break;
	case 3042: typestring = "River/stream, centerline in water body, perennial, length 20-< 25 km, or l2-< l6 mi"; break;
	case 3043: typestring = "River/stream, centerline in water body, perennial, length 25-< 30 km, or l6-< l9 mi"; break;
	case 3044: typestring = "River/stream, centerline in water body, perennial, length 30-< 40 km, or l9-< 25 mi"; break;
	case 3045: typestring = "River/stream, centerline in water body, perennial, length 40-< 50 km, or 25-< 3l mi"; break;
	case 3046: typestring = "River/stream, centerline in water body, perennial, length 50-< 60 km, or 3l-< 37 mi"; break;
	case 3047: typestring = "River/stream, centerline in water body, perennial, length 60-< 80 km, or 37-< 50 mi"; break;
	case 3048: typestring = "River/stream, centerline in water body, perennial, length 80+ km, or 50+ mi"; break;
	case 3050: typestring = "River/stream, centerline in water body, intermittent, length < 2 km, or < l mi"; break;
	case 3051: typestring = "River/stream, centerline in water body, intermittent, length 2-< 4 km, or l-< 2 mi"; break;
	case 3052: typestring = "River/stream, centerline in water body, intermittent, length 4-< 6 km, or 2-< 4 mi"; break;
	case 3053: typestring = "River/stream, centerline in water body, intermittent, length 5-< 8 km, or 4-< 5 mi"; break;
	case 3054: typestring = "River/stream, centerline in water body, intermittent, length 8-< l0 km, or 5-< 6 mi"; break;
	case 3055: typestring = "River/stream, centerline in water body, intermittent, length l0-< l5 km, or 6-< 9 mi"; break;
	case 3056: typestring = "River/stream, centerline in water body, intermittent, length l5-< 20 km, or 9-< l2 mi"; break;
	case 3057: typestring = "River/stream, centerline in water body, intermittent, length 20-< 25 km, or l2-< l6 mi"; break;
	case 3058: typestring = "River/stream, centerline in water body, intermittent, length 25-< 30 km, or l6-< l9 mi"; break;
	case 3059: typestring = "River/stream, centerline in water body, intermittent, length 30+ km, or l9+ mi"; break;
	case 3060: typestring = "Braided stream, average width of braid 6+ km, or 4+ mi"; break;
	case 3061: typestring = "Braided stream, average width of braid 0-< 2 km, or 0-< l mi"; break;
	case 3062: typestring = "Braided stream, average width of braid 2-< 4 km, or l-< 2 mi"; break;
	case 3063: typestring = "Braided stream, average width of braid 4-< 6 km, or 2-< 4 mi"; break;
	case 3070: typestring = "Canal, navigable, length < l km, or < l mi"; break;
	case 3071: typestring = "Canal, navigable, length l-< l0 km, or l-< 6 mi"; break;
	case 3072: typestring = "Canal, navigable, length l0-< 20 km, or 6-< l2 mi"; break;
	case 3073: typestring = "Canal, navigable, length 20-< 40 km, or l2-< 25 mi"; break;
	case 3074: typestring = "Canal, navigable, length 40-< 60 km, or 25-< 37 mi"; break;
	case 3075: typestring = "Canal, navigable, length 60-< 80 km, or 37-< 50 mi"; break;
	case 3076: typestring = "Canal, navigable, length 80+ km, or 50+ mi"; break;
	case 3077: typestring = "Canal, other, length < l km, or < l mi"; break;
	case 3078: typestring = "Canal, other, length l-< l0 km, or l-< 6 mi"; break;
	case 3079: typestring = "Canal, other, length l0-< 20 km, or 6-< l2 mi"; break;
	case 3080: typestring = "Canal, other, length 20-< 40 km, or l2-< 25 mi"; break;
	case 3081: typestring = "Canal, other, length 40-< 60 km, or 25-< 37 mi"; break;
	case 3082: typestring = "Canal, other, length 60-< 80 km, or 37-< 50 mi"; break;
	case 3083: typestring = "Canal, other, length 80+ km, or 50+ mi"; break;
	case 3086: typestring = "Ditch (perennial)"; break;
	case 3095: typestring = "Intercoastal waterway"; break;
	case 4000: typestring = "U.S. coastline including Great Lakes     "; break;
	case 4001: typestring = "Perennial water body, lake, reservoir, and island, length < 2 km, or < l mi"; break;
	case 4002: typestring = "Perennial water body, lake, reservoir, and island, length 2-< 4 km, or l-< 2 mi"; break;
	case 4003: typestring = "Perennial water body, lake, reservoir, and island, length 4-< 6 km, or 2-< 4 mi"; break;
	case 4004: typestring = "Perennial water body, lake, reservoir, and island, length 6-< 8 km, or 4-< 5 mi"; break;
	case 4005: typestring = "Perennial water body, lake, reservoir, and island, length 8-< l0 km, or 5-< 6 mi"; break;
	case 4006: typestring = "Perennial water body, lake, reservoir, and island, length l0-< l5 km, or 16-< 9 mi"; break;
	case 4007: typestring = "Perennial water body, lake, reservoir, and island, length l5-< 20 km, or 9-< l2 mi"; break;
	case 4008: typestring = "Perennial water body, lake, reservoir, and island, length 20-< 25 km, or l2-< l6 mi"; break;
	case 4009: typestring = "Perennial water body, lake, reservoir, and island, length 25-< 30 km, or l6-< l9 mi"; break;
	case 4010: typestring = "Perennial water body, lake, reservoir, and island, length 30-< 40 km, or l9-< 25 mi"; break;
	case 4011: typestring = "Perennial water body, lake, reservoir, and island, length 40-< 50 km, or 25-< 3l mi"; break;
	case 4012: typestring = "Perennial water body, lake, reservoir, and island, length 50-< 60 km, or 3l-< 37 mi"; break;
	case 4013: typestring = "Perennial water body, lake, reservoir, and island, length 60-< 80 km, or 37-< 50 mi"; break;
	case 4014: typestring = "Perennial water body, lake, reservoir, and island, length 80+ km, or 50+ mi"; break;
	case 4021: typestring = "Intermittent water body, lake or reservoir, length < 2 km, or < l mi"; break;
	case 4022: typestring = "Intermittent water body, lake or reservoir, length 2-< 4 km, or l-< 2 mi"; break;
	case 4023: typestring = "Intermittent water body, lake or reservoir, length 4-< 6 km, or 2-< 4 mi"; break;
	case 4024: typestring = "Intermittent water body, lake or reservoir, length 6-< 8 km, or 4-< 5 mi"; break;
	case 4025: typestring = "Intermittent water body, lake or reservoir, length 8-< l0 km, or 5-< 6 mi"; break;
	case 4026: typestring = "Intermittent water body, lake or reservoir, length l0-< l5 km, or 6-< 9 mi"; break;
	case 4027: typestring = "Intermittent water body, lake or reservoir, length l5-< 20 km, or 9-< l2 mi"; break;
	case 4028: typestring = "Intermittent water body, lake or reservoir, length 20-< 25 km, or l2-< l6 mi"; break;
	case 4029: typestring = "Intermittent water body, lake or reservoir, length 25-< 30 km, or l6-< l9 mi"; break;
	case 4030: typestring = "Intermittent water body, lake or reservoir, length 30-< 40 km, or l9-< 25 mi"; break;
	case 4031: typestring = "Intermittent water body, lake or reservoir, length 40-< 50 km, or 25-< 3l mi"; break;
	case 4032: typestring = "Intermittent water body, lake or reservoir, length 50-< 60 km, or 3l-< 37 mi"; break;
	case 4033: typestring = "Intermittent water body, lake or reservoir, length 60-< 80 km, or 37-< 50 mi"; break;
	case 4034: typestring = "Intermittent water body, lake or reservoir, length 80+ km, or 50+ mi"; break;
	case 4040: typestring = "Marsh/swamp and salt marsh, length < l0 km, or < 6 mi"; break;
	case 4041: typestring = "Marsh/swamp and salt marsh, length l0-< l7 km, or 6-< ll mi"; break;
	case 4042: typestring = "Marsh/swamp and salt marsh, length l7-< 25 km, or ll-< l6 mi"; break;
	case 4043: typestring = "Marsh/swamp and salt marsh, length 25-< 37 km, or l6-< 23 mi"; break;
	case 4044: typestring = "Marsh/swamp and salt marsh, length 37-< 50 km, or 23-< 3l mi"; break;
	case 4045: typestring = "Marsh/swamp and salt marsh, length 50+ km, or 3l+ mi"; break;
	case 4050: typestring = "Dry lake and alkali flat, length < 2 km, or < l mi"; break;
	case 4051: typestring = "Dry lake and alkali flat, length 2-< 4 km, or l-< 2 mi"; break;
	case 4052: typestring = "Dry lake and alkali flat, length 4-< 6 km, or 2-< 4 mi"; break;
	case 4053: typestring = "Dry lake and alkali flat, length 6+ km, or 4+ mi"; break;
	case 4060: typestring = "Glacier, length < 4 km, or < 2 mi"; break;
	case 4061: typestring = "Glacier, length 4-< l0 km, or 2-< 6 mi"; break;
	case 4062: typestring = "Glacier, length l0-< l7 km, or 6-< ll mi"; break;
	case 4063: typestring = "Glacier, length l7-< 25 km, or ll-< l6 mi"; break;
	case 4064: typestring = "Glacier, length 25-< 37 km, or l6-< 23 mi"; break;
	case 4065: typestring = "Glacier, length 37-< 50 km, or 23-< 3l mi"; break;
	case 4066: typestring = "Glacier, length 50+ km, or 3l+ mi"; break;
	case 5001: typestring = "Interstate"; break;
	case 5002: typestring = "Major U.S., limited access, divided"; break;
	case 5003: typestring = "Major State, limited access, divided"; break;
	case 5004: typestring = "Major other, limited access, divided"; break;
	case 5005: typestring = "Toll road1"; break;
	case 5006: typestring = "Interstate connector1"; break;
	case 5007: typestring = "Limited access, divided connector1"; break;
	case 5008: typestring = "Toll connector1"; break;
	case 5009: typestring = "Interstate, under construction"; break;
	case 5010: typestring = "Interstate, proposed"; break;
	case 5013: typestring = "Minor U.S., limited access, 3l0 km (500 mi) and longer"; break;
	case 5014: typestring = "U.S. non-limited access, 3l0 km (500 mi) and longer"; break;
	case 5015: typestring = "Minor U.S. limited access, less than 3l0 km (500 mi)"; break;
	case 5016: typestring = "U.S. non-limited access, less than 3l0 km (500 mi)"; break;
	case 5017: typestring = "Other minor U.S. limited access"; break;
	case 5018: typestring = "Other U.S.2"; break;
	case 5019: typestring = "Other minor State primary, limited access"; break;
	case 5020: typestring = "Other State primary"; break;
	case 5021: typestring = "Minor U.S. parallel, within l0 km (6 mi)"; break;
	case 5022: typestring = "U.S. parallel, within l0 km (6 mi)"; break;
	case 5023: typestring = "Minor State parallel, within l0 km (6 mi)"; break;
	case 5024: typestring = "State parallel, within l0 km (6 mi)"; break;
	case 5028: typestring = "State secondary (all weather, hard surface)"; break;
	case 5031: typestring = "Light duty (all weather, improved)"; break;
	case 5041: typestring = "Unimproved (fair or dry weather)"; break;
	case 5061: typestring = "Tunnel, road"; break;
	case 5062: typestring = "Ferry, auto"; break;
	case 5071: typestring = "Class 1, category A, main line"; break;
	case 5072: typestring = "Class 1, category B, main line"; break;
	case 5073: typestring = "Class 1, category A, branch line"; break;
	case 5074: typestring = "Class 1, category B, branch line"; break;
	case 5075: typestring = "Other railroad"; break;
	case 5078: typestring = "Tunnel, railroad"; break;
	case 5079: typestring = "Ferry, railroad"; break;
	case 5080: typestring = "Class 1, category A, main-line connector1"; break;
	case 6000: typestring = "International treaty line"; break;
	case 6001: typestring = "National (land)"; break;
	case 6002: typestring = "National (water)"; break;
	case 6005: typestring = "State/provincial (land)"; break;
	case 6006: typestring = "State/provincial (water)"; break;
	case 6009: typestring = "County, parish, Alaskan borough, or large independent city (land)"; break;
	case 6010: typestring = "County, parish, Alaskan borough, or large independent city (water)"; break;
	case 60ll: typestring = "Corporate limit (1 million and over population)"; break;
	case 6012: typestring = "Corporate limit (1/2 to less than l million population)"; break;
	case 6014: typestring = "Small independent city (usually not shown as a county)"; break;
	case 6021: typestring = "National park, length at longest dimension 0-< 2 km, or 0-< l mi"; break;
	case 6022: typestring = "National park, length at longest dimension 2-< 8 km, or l-< 5 mi"; break;
	case 6023: typestring = "National park, length at longest dimension 8-< l4 km, or 5-< 9 mi"; break;
	case 6024: typestring = "National park, length at longest dimension 4-< 20 km, or 9-< 12 mi"; break;
	case 6025: typestring = "National park, length at longest dimension 20+ km, or 12+ mi"; break;
	case 6026: typestring = "National monument, length at longest dimension 0-< 2 km, or 0-< 1 mi"; break;
	case 6027: typestring = "National monument, length at longest dimension 2-< 8 km, or 1-< 5 mi"; break;
	case 6028: typestring = "National monument, length at longest dimension 8-< 14 km, or 5-< 9 mi"; break;
	case 6029: typestring = "National monument, length at longest dimension 14-< 20 km, or 9-< 12 mi"; break;
	case 6030: typestring = "National monument, length at longest dimension 20+ km, or 12+ mi"; break;
	case 6031: typestring = "National seashore or lakeshore, length at longest dimension 0-< 2 km, or 0-< 1 mi"; break;
	case 6032: typestring = "National seashore or lakeshore, length at longest dimension 2-< 8 km, or 1-< 5 mi"; break;
	case 6033: typestring = "National seashore or lakeshore, length at longest dimension 8-< 14 km, or 5-< 9 mi"; break;
	case 6034: typestring = "National seashore or lakeshore, length at longest dimension 4-< 20 km, or 9-< 12 mi"; break;
	case 6035: typestring = "National seashore or lakeshore, length at longest dimension 20+ km, or 12+ mi"; break;
	case 6036: typestring = "National recreation area, length at longest dimension 0-< 2 km, or 0-< 1 mi"; break;
	case 6037: typestring = "National recreation area, length at longest dimension 2-< 8 km, or 1-< 5 mi"; break;
	case 6038: typestring = "National recreation area, length at longest dimension 8-< 14 km, or 5-< 9 mi"; break;
	case 6039: typestring = "National recreation area, length at longest dimension 14-< 20 km, or 9-< 12 mi"; break;
	case 6040: typestring = "National recreation area, length at longest dimension 20+ km, or 12+ mi"; break;
	case 6041: typestring = "National wilderness area, length at longest dimension 0-< 2 km, or 0-< 1 mi"; break;
	case 6042: typestring = "National wilderness area, length at longest dimension 2-< 8 km, or 1-< 5 mi"; break;
	case 6043: typestring = "National wilderness area, length at longest dimension 8-< 14 km, or 5-< 9 mi"; break;
	case 6044: typestring = "National wilderness area, length at longest dimension 14-< 20 km, or 9-< 12 mi"; break;
	case 6045: typestring = "National wilderness area, length at longest dimension 20+ km, or 12+ mi"; break;
	case 6051: typestring = "National forest, length at longest dimension 0-< 2 km, or 0-< 1 mi"; break;
	case 6052: typestring = "National forest, length at longest dimension 2-< 8 km, or 1-< 5 mi"; break;
	case 6053: typestring = "National forest, length at longest dimension 8-< l4 km, or 5-< 9 mi"; break;
	case 6054: typestring = "National forest, length at longest dimension l4-< 20 km, or 9-< l2 mi"; break;
	case 6055: typestring = "National forest, length at longest dimension 20+ km, or l2+ mi"; break;
	case 6056: typestring = "National grassland, length at longest dimension 0-< 2 km, or 0-< 1 mi"; break;
	case 6057: typestring = "National grassland, length at longest dimension 2-< 8 km, or l-< 5 mi"; break;
	case 6058: typestring = "National grassland, length at longest dimension 8-< 14 km, or 5-< 9 mi"; break;
	case 6059: typestring = "National grassland, length at longest dimension 14-< 20 km, or 9-< 12 mi"; break;
	case 6060: typestring = "National grassland, length at longest dimension 20+ km, or l2+ mi"; break;
	case 6061: typestring = "National wildlife refuge, length at longest dimension 0-< 2 km, or 0-< l mi"; break;
	case 6062: typestring = "National wildlife refuge, length at longest dimension 2-< 8 km, or l-< 5 mi"; break;
	case 6063: typestring = "National wildlife refuge, length at longest dimension 8-< l4 km, or 5-< 9 mi"; break;
	case 6064: typestring = "National wildlife refuge, length at longest dimension l4-< 20 km, or 9-< l2 mi"; break;
	case 6065: typestring = "National wildlife refuge, length at longest dimension 20+ km, or l2+ mi"; break;
	case 6066: typestring = "Federal Indian reservation, length at longest dimension 0-< 2 km, or 0-< l mi"; break;
	case 6067: typestring = "Federal Indian reservation, length at longest dimension 2-< 8 km, or l-< 5 mi"; break;
	case 6068: typestring = "Federal Indian reservation, length at longest dimension 8-< l4 km, or 5-< 9 mi"; break;
	case 6069: typestring = "Federal Indian reservation, length at longest dimension l4-< 20 km, or 9-< l2 mi"; break;
	case 6070: typestring = "Federal Indian reservation, length at longest dimension 20+ km, or l2+ mi"; break;
	case 6081: typestring = "Federal Military reservation, areas of l-< 405 ha, or l-< l000 acres"; break;
	case 6082: typestring = "Federal Military reservation, areas of 405+ ha, or l000+ acres"; break;
	case 6087: typestring = "National park, closure line"; break;
	case 6088: typestring = "National monument, closure line"; break;
	case 6089: typestring = "National seashore or lakeshore, closure line"; break;
	case 6090: typestring = "National recreation area, closure line"; break;
	case 6091: typestring = "National wilderness area, closure line"; break;
	case 6092: typestring = "National forest, closure line"; break;
	case 6093: typestring = "National grassland, closure line"; break;
	case 6094: typestring = "National wildlife refuge, closure line"; break;
	case 6095: typestring = "Indian reservation, closure line"; break;
	case 6097: typestring = "Military reservation, closure line"; break;
	case 7001: typestring = "Commmercial airfield"; break;
	case 7002: typestring = "Military airfield"; break;
	case 7020: typestring = "Alaska pipeline"; break;
	default:
	    sprintf(buf, "Unknown 290 series feature %d", minor_attribute());
	    typestring = buf;
	    break;
	}
	break;
    default:
	sprintf(buf, "Unknown DLG major code %d", major_attribute());
	typestring = buf;
	break;
    }
    sprintf(buffer, "%s, %s", MapFeature::describe(), typestring);
    return buffer;
}
